/* global AFRAME, THREE */

/**
 * Component that controlls the cleaning robot (cBot) entity. In essence, in each step (frame), it operates like this:
 * 1. Check what should be done, i.e. clean a position or go to a parking position.
 * 2. Check position of other citizens in the scene.
 * 3. Call the verified controller (window.scCBotController.updateAndEvaluate) to decide what to do, without endangering other citizens (users).
 * 4. Update the cBot position, state and colors according to the output of the verified controller. Aslo update the dirty positions.
 */
AFRAME.registerComponent("cbot-master-controller", {
    schema: {
      cBotState: {type: "int", default: 0},                //0 = cBot is off                  (botOn==false)
                                                           //1 = cBot is on  and     cleaning (botOn==true && botCleaning==true)
                                                           //2 = cBot is on  and not cleaning (botOn==true && botCleaning==false)
                                                           //(botOn and botCleaning are from CBotController.updateAndEvaluate)
      MOVEMENT_SPEED: {type: "number", default: 0.02},      //cBot MOVEMENT_SPEED, in meters per frame
      PARKING_POSITION:{type: "vec2", default: {x:0, y:0}}, //the position in the world where the cBot starts and ends its journey (y is z in 3D)
                                                           //(horizontal plane coordinates only)
      TICKS_TO_RESTART: {type: "int", default: 500}
    },

    init: function () {
      
      /*
      //code to try movement control (can be used before inclucing scCBotController)
      this.counter=3000;
      this.rotMul=0;
      */
      
      
      /* 8.
      this.actualTicks2Restart = this.data.TICKS_TO_RESTART;
      8. */
      
      //change this to true, if you want to see some position logs in the first frame with the cBot
      this.logNeeded=false;
      
      
      /* 5.
      //Create a collection of connected users (citizens)
      //getElementsByClassName cleates a live collection, so we do not need to care when a user dis/connects
      this.citizens=document.getElementsByClassName("avatarCitizen"); 
      5. */
      
      /* 9.
      //get access to the head of the cBot
      this.head = document.querySelector(".cbot-head");
      9. */
      
      /* 1.
      //put cBot to the PARKING_POSITION, facing north
      this.PARKING_POSITION=new THREE.Vector3(this.data.PARKING_POSITION.x, 1, this.data.PARKING_POSITION.y);
      this.el.object3D.position.set(this.PARKING_POSITION.x, this.PARKING_POSITION.y, this.PARKING_POSITION.z);
      this.el.object3D.rotation.set(0, 0, 0);
      1. */
      
      /* 2.
      //fill in a collection of positions the cBot has to clean and load the first position
      this.dirtyPositions = document.querySelectorAll("[cbot-dirty-position]"); // this collection is not a live collection (it's OK, we do not expect it to change while cBot is active)
      this.noOfDirtyPositions=this.dirtyPositions.length;
      this.nextDirtyPositionIndex=0;
      this.loadNextPosition2Clean();
      2. */
      
      //initialize other variables to be used in the tick function
      this.instructions4CBot=[false, false, 0, 0];//[botOn, botCleaning, angleInMin, dst]
      this.vector2Target = new THREE.Vector3();
      this.targetAngleInMin=0;
      this.targetDistanceInMm=0;
      this.sensorArrayReadings=[30000,30000,30000,30000,30000,30000,30000,30000];
      this.newCBotState=this.data.cBotState;
      

      
    },

    tick: function () {
      
      /*
      //code to try movement control (can be used before inclucing scCBotController)
      if(this.counter>0){
        if(this.counter%500==0){
          this.el.object3D.rotation.y = 1.0*this.rotMul;
          console.log("rot, this.counter= "+this.counter);
          this.rotMul++;
        }
        this.el.object3D.translateZ(-this.data.MOVEMENT_SPEED);
        this.counter--;
      }
      */      
      
      if(window.scCBotController){  //the verified controller is a module, so it is loaded after this component. Therefore, we wait for it a couple of frames
        
        if(this.position2Clean){ //if there is a position to clean, try to go to it and clean it
        
          /* 6.
          //Get readings from the sensor array (i.e. positions of other users)
          this.readSensorArray();
          6. */
          
          /* 3.
          //Get the distance and angle to the position to clean from the current position of the cBot
          this.updateDistanceAndAngleToTarget(this.position2Clean.object3D.position);
          
          //writes the distance and angle computed to the console (only in the first frame with cBot)
          if(this.logNeeded){           
            this.logPositions2Console();
            this.logNeeded=false;            
          }
          3. */
          
          /* 4.
          //calls the verified controller to check whether it is safe to perform the desired operation (i.e. cleaning)
          this.instructions4CBot = //[botOn, botCleaning, angleInMin, dst]
            window.scCBotController.updateAndEvaluate(1,this.targetAngleInMin,this.targetDistanceInMm,...this.sensorArrayReadings);
          
          //process and execute the instructions for the cBot, returned by the verified controller
          this.newCBotState=this.calculateCBotState(this.instructions4CBot[0],this.instructions4CBot[1]);          
          if(this.newCBotState===2){//cBot is on and not cleaning
            this.moveCBotTowardsTarget();
          }else if(this.newCBotState===1){//cBot is on and cleaning
            this.position2CleanAFrameComponent.doCleaningStep(1);              
            if(this.position2CleanAFrameComponent.isClean()){
              this.loadNextPosition2Clean();
            }
          }//otherwise, the cBot is turned off, so do nothing
          4. */
          
        }
        /* 7.
        else
        {  //if all positions have been cleaned

          
          if(this.data.cBotState===0){//if the cBot has been turned off
          
        7. */    
            /* 8.
            //restart the cBot after this.data.TICKS_TO_RESTART ticks (frames)
            if(this.actualTicks2Restart>0){
              this.actualTicks2Restart--;
            }else{
              this.actualTicks2Restart = this.data.TICKS_TO_RESTART;
              this.makePositionsDirty();
              this.loadNextPosition2Clean();
            }
            8. */
            
          /* 7.
          }else{//if the cBot is on,
            
            
            //cBot should go to the parking position
            
            //Get readings from the sensor array (i.e. positions of other users)
            this.readSensorArray();

            //Get the distance and angle to the position to clean from the current position of the cBot
            this.updateDistanceAndAngleToTarget(this.PARKING_POSITION);

            this.instructions4CBot = //[botOn, botCleaning, angleInMin, dst]
              window.scCBotController.updateAndEvaluate(3,this.targetAngleInMin,this.targetDistanceInMm,...this.sensorArrayReadings);

            this.newCBotState=this.calculateCBotState(this.instructions4CBot[0],this.instructions4CBot[1]);
            
            

            if(this.newCBotState===2){//cBot is on and not cleaning
              this.moveCBotTowardsTarget();
            }
            
            

          }
          
                    
        }
        7. */
        
        if(this.newCBotState!=this.data.cBotState){
          this.el.setAttribute("cbot-master-controller", "cBotState", this.newCBotState);
        }

        
      }
      

      
    },
  
    /* 9.
    update: function () {
      if(this.head){
      let cBotheadColor="grey";
      if (this.data.cBotState===1){ //cBot is on and cleaning
        cBotheadColor="red";
      }else if (this.data.cBotState===2){//cBot is on and not cleaning 
        cBotheadColor="darkorange";        
      }
      this.head.setAttribute("material", "color", cBotheadColor);
        
      }
    },
    9. */
  
    /**
     * Loads the next position to clean element to this.position2Clean 
     * and its cbot-dirty-position component to position2CleanAFrameComponent
     */
    loadNextPosition2Clean: function () {
      if(this.nextDirtyPositionIndex<this.noOfDirtyPositions){
        this.position2Clean=this.dirtyPositions[this.nextDirtyPositionIndex];
        this.position2CleanAFrameComponent = this.position2Clean.components["cbot-dirty-position"];
        this.nextDirtyPositionIndex++;
      }else{
        this.position2Clean = undefined;
        this.position2CleanAFrameComponent = undefined;
      }
    },

    /**
     * Gets readings from the sensor array (i.e. positions of other users)
     * to this.sensorArrayReadings
     */
    readSensorArray: function () {
      //first, set distances for all sensors to maximum values
      this.sensorArrayReadings=[30000,30000,30000,30000,30000,30000,30000,30000]; 
      
      let cBotPosition = this.el.object3D.getWorldPosition(new THREE.Vector3());
      let citizenAngleInMin=0;
      let citizenDistanceInMm=0;
      let vector2Citizen=new THREE.Vector3();
      let arrayIndex=-1;
      if(this.citizens.length>0){
        for (const citizen of this.citizens) {
          vector2Citizen=vector2Citizen.subVectors(citizen.object3D.position,cBotPosition);
          if (!this.isVector3Zero(vector2Citizen)) {
            citizenAngleInMin=this.cBotAngle2TargetInMin(vector2Citizen);
            citizenDistanceInMm=this.cBotDistance2TargetInMm(vector2Citizen);            
          }
          if(citizenDistanceInMm<=30000){
            arrayIndex=(Math.trunc(citizenAngleInMin/2700)+7)%8;
            if(this.sensorArrayReadings[arrayIndex]>citizenDistanceInMm){
              this.sensorArrayReadings[arrayIndex]=citizenDistanceInMm;
            }
          }
        }
      }
    },
  
    /**
     * Calculates the distance and angle to the targetPosition from the current position of the cBot
     * and stores them to this.targetAngleInMin and  this.targetDistanceInMm.
     * Also stores the directional vector between these two positions to this.vector2Target
     * @param targetPosition - vector (THREE.Vector3) with the position to go
     */    
    updateDistanceAndAngleToTarget: function (targetPosition) {
      this.vector2Target = this.vector2Target.subVectors(targetPosition, this.el.object3D.getWorldPosition(new THREE.Vector3()));
      this.targetAngleInMin=0;
      this.targetDistanceInMm=0;
      if (!this.isVector3Zero(this.vector2Target)) {
        this.targetAngleInMin=this.cBotAngle2TargetInMin(this.vector2Target);
        this.targetDistanceInMm=this.cBotDistance2TargetInMm(this.vector2Target);            
      }
    },

    /**
     * Moves (and rotates) the cBot towards its target according to this.instructions4CBot and using this.data.MOVEMENT_SPEED
     */
    moveCBotTowardsTarget: function (){
      //convert the distance from mm to m          
      this.instructions4CBot[3]=this.instructions4CBot[3]/1000; 

      //rotate cBot
      this.el.object3D.rotation.y=this.convertCBotAngle2AFrameAngle(this.instructions4CBot[2]);

      //move cBot (only when needed, i.e. dst>0)
      if(this.instructions4CBot[3]>0){
        if(this.instructions4CBot[3]<this.data.MOVEMENT_SPEED){
          this.el.object3D.translateZ(-this.instructions4CBot[3]); //move 0.3m forward              
        }else{
          this.el.object3D.translateZ(-this.data.MOVEMENT_SPEED); //move 0.3m forward            
        }
      }
    },
  
  
    /**
     *makes all the positions dirty (again)
     */
    makePositionsDirty: function (){
      for (let position of this.dirtyPositions) {
        position.components["cbot-dirty-position"].makeDirty();
      }
      this.nextDirtyPositionIndex=0;
    },
  
    /**
     *Writes the cBot position, the position to clean and  distance and angle between them to the console
     */
    logPositions2Console: function (){
      console.log("cBot position:");
      console.log(this.el.object3D.getWorldPosition(new THREE.Vector3()));
      console.log("Position to clean:");
      console.log(this.position2Clean.object3D.position);  
      console.log("this.vector2Target:");
      console.log(this.vector2Target);
      console.log("this.targetAngleInMin="+this.targetAngleInMin);
      console.log("this.targetDistanceInMm="+this.targetDistanceInMm);
    },

  
    /**
    * Calculates cBot state according to values returned by CBotController.updateAndEvaluate
    * @param botOn - true if cBot is turned on, false otherwise
    * @param botCleaning - true if cBot is cleaning, false otherwise
    * @returns 0 (cBot is off) if botOn===false
    *          1 (cBot is on  and     cleaning) if botOn===true && botCleaning===true
    *          2 (cBot is on  and not cleaning) if botOn===true && botCleaning===false
    */
    calculateCBotState: function (botOn, botCleaning) {
      if(botOn===false){
        return 0;
      }else if (botOn===true){
        if(botCleaning===true){
          return 1;
        }else if (botCleaning===false){
          return 2;
        }else{
          return 0;
        }
      }else{
        return 0;
      }
      
    },
  
  
    /**
     * calculates the cBot Angle (in minutes, clockwise) of vector2Target (THREE.Vector3)
     */
    cBotAngle2TargetInMin: function (vector2Target) {
      
      //angleDgClkwsMin2Pl180 = angle in degrees (float), clockwise, from -180dgr. to 180 dgr.
      let angleDgClkwsMin2Pl180=(Math.atan2(vector2Target.x, -vector2Target.z))* 180 / Math.PI;
      let angleDegClockwise = angleDgClkwsMin2Pl180>=0 ? angleDgClkwsMin2Pl180 : 360+angleDgClkwsMin2Pl180;
      if(this.logNeeded){
         console.log("angleDgClkwsMin2Pl180="+angleDgClkwsMin2Pl180);
         console.log("angleDegClockwise="+angleDegClockwise+" min "+ Math.round(angleDegClockwise*60));
      }
      return Math.round(angleDegClockwise*60);
    },

  
  
    /**
     * calculates the cBot distance (in mm) of vector2Target (THREE.Vector3) with y-coordinate (height) ignored
     */
    cBotDistance2TargetInMm: function (vector2Target) {
      vector2Target.y=0;
      return Math.round(vector2Target.length()*1000);
    },
  
    /**
     * converts a cBot Angle (in minutes, clockwise) to A-Frame angle (in radians, counter-clockwise)
     */
    convertCBotAngle2AFrameAngle(cBotAngleInMin){
      return THREE.Math.degToRad((21600-cBotAngleInMin)/60);
    },

    /**
     * Finds out whether a THREE.Vector3 vector is a zero vector
     * @param vector - a THREE.Vector3 vector to be examined
     * @returns true if vector is a zero vector, false otherwise
     */
    isVector3Zero: function (vector) {
        return vector.x === 0 && vector.y === 0 && vector.z === 0;
    }
  

});

